@@ -8,23 +8,26 @@ In The Eleventh International Conference on Learning Representations. 2023
 Agents in the HotPotQA environment can perform searches and lookups to retrieve specific information.
 The environment supports 3 tools:
 
-Search[entity]: Search for a specific entity on Wikipedia and return relevant information.
-Lookup[keyword]: Find and return the next sentence containing the keyword in the current passage.
-Finish[answer]: Submit the final answer to the question and conclude the task.
+search[entity]: Search for a specific entity on Wikipedia and return relevant information.
+lookup[keyword]: Find and return the next sentence containing the keyword in the current passage.
+submit_answer[answer]: Submit the final answer to the question and conclude the task.
 """
 
 import logging
+import os
 import random
 import re
 import string
 from collections.abc import Callable
 from enum import StrEnum
-from typing import Any, ClassVar, cast
+from typing import TYPE_CHECKING, Any, ClassVar, cast
+from uuid import UUID
 
 import httpx
+import httpx_aiohttp
 from bs4 import BeautifulSoup
 from datasets import load_dataset
-from pydantic import BaseModel, ConfigDict, Field
+from pydantic import BaseModel, ConfigDict, Field, field_validator
 from tenacity import retry, stop_after_attempt, wait_exponential_jitter
 
 from aviary.core import (
@@ -32,18 +35,21 @@ from aviary.core import (
     EvalAnswerMode,
     Frame,
     Message,
+    Messages,
     TaskDataset,
     Tool,
     ToolRequestMessage,
-    ToolResponseMessage,
     eval_answer,
 )
 
+if TYPE_CHECKING:
+    from datasets import Dataset
+
 logger = logging.getLogger(__name__)
 
 
 # Jitter in case we have lots of requests at once, to space them out slightly
-@retry(stop=stop_after_attempt(2), wait=wait_exponential_jitter(initial=1, max=4))
+@retry(stop=stop_after_attempt(2), wait=wait_exponential_jitter(max=4))
 async def fetch_with_retry(
     client: httpx.AsyncClient, url: str, **kwargs
 ) -> httpx.Response:
@@ -72,8 +78,10 @@ class HotPotQAEnvState(BaseModel):
     )
     last_action_is_lookup: bool = Field(
         default=False,
-        description="Whether the last action was a lookup action."
-        "Default is False, as after reset the agent has not yet taken any action.",
+        description=(
+            "Whether the last action was a lookup action."
+            "Default is False, as after reset the agent has not yet taken any action."
+        ),
     )
     last_lookup: str | None = Field(
         default=None, description="The last lookup keyword."
@@ -86,8 +94,6 @@ class HotPotQAEnvState(BaseModel):
     )
     page: str | None = Field(default=None, description="The current Wikipedia page.")
 
-    evaluation_mode: EvalAnswerMode = EvalAnswerMode.CONTAINS
-
 
 def create_tool(function: Callable, name: str) -> Tool:
     """Create a Tool object from a function and set its name.
@@ -181,6 +187,7 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
 
     def __init__(
         self,
+        question_id: UUID | None,
         question: str,
         correct_answer: str | float,
         correct_reward: float = 1.0,
@@ -190,6 +197,7 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
         proxy: str | None = None,
     ):
         super().__init__()
+        self.question_id = question_id
         self.question = question
         # Normalize the correct answer once here as a minor performance optimization
         self.normalized_correct_answer = normalize_answer(correct_answer)
@@ -208,14 +216,14 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
 
         # Title case tool names to match third party demonstration data
         self.tools = [
-            create_tool(self.search, "Search"),
-            create_tool(self.construct_lookup_list, "Lookup"),
-            create_tool(self.finish, "Finish"),
+            Tool.from_function(self.search),
+            Tool.from_function(self.lookup),
+            Tool.from_function(self.submit_answer),
         ]
 
     @classmethod
     def from_task(cls, task: str) -> "HotPotQAEnv":
-        return cls(question=task, correct_answer=0.0)
+        return cls(question_id=None, question=task, correct_answer=0.0)
 
     async def calculate_answer_reward(self, answer: str | None) -> float:
         """Calculate the reward based on the agent's answer.
@@ -238,7 +246,7 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
             else self.incorrect_reward
         )
 
-    async def reset(self) -> tuple[list[Message], list[Tool]]:
+    async def reset(self) -> tuple[Messages, list[Tool]]:
         """Reset the HotPotQA environment to an initial state.
 
         This method resets the environment to its initial state, setting up necessary variables and tools
@@ -247,23 +255,30 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
 
         Returns:
             tuple: A tuple containing:
-                - list[Message]: The initial observation wrapped in a Message object.
-                - list[Tool]: A list of tools (Search, Lookup, and Finish) available for the agent.
+                - Messages: The initial observation wrapped in a Message object.
+                - list[Tool]: A list of tools (search, lookup, and submit_answer) available for the agent.
 
         Example:
-            >>> env = HotPotQAEnv()
-            >>> initial_obs, tools = env.reset(seed=42, idx=5)
-            >>> print(initial_obs)
-            [Message(content='Question: <question_text>')]
-            >>> print(tools)
-            [<Tool: Search>, <Tool: Lookup>, <Tool: Finish>]
+            >>> env = HotPotQAEnv(
+            ...     question_id=UUID("00000000-5a8d-7341-5542-99441c6b9fe5"),
+            ...     question=(
+            ...         "Musician and satirist Allie Goertz wrote a song about the"
+            ...         ' "The Simpsons" character Milhouse, who Matt Groening named after who?'
+            ...     ),
+            ...     correct_answer="President Richard Nixon",
+            ... )
+            >>> obs, tools = await env.reset()
+            >>> print(obs)
+            [Message(role='user', content='Question: Musician and satirist Allie Goertz wrote a song about the "The Simpsons" character Milhouse, who Matt Groening named after who?')]
+            >>> print([t.info.name for t in tools])
+            ['search', 'lookup', 'submit_answer']
         """
         self.state = self.State()
         return [Message(content=f"Question: {self.question}")], self.tools
 
     async def step(
         self, action: ToolRequestMessage
-    ) -> tuple[list[Message], float, bool, bool]:
+    ) -> tuple[Messages, float, bool, bool]:
         """Take a step in the environment. Assume only one tool at a time can be called for HotpotQA.
 
         This method processes an action message, which can be a tool request, a finish request, or an error message.
@@ -273,19 +288,22 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
             action: Action to take.
 
         Returns:
-            Tuple[List[ToolResponseMessage], bool]: A tuple containing:
-                - list[ToolResponseMessage]: The response message(s) from the executed tool.
+            Tuple[Messages, bool]: A tuple containing:
+                - Messages: The response message(s) from the executed tool.
                 - bool: The done status indicating whether the episode is finished.
 
         Example:
-            >>> env = HotPotQAEnv()
+            >>> from aviary.core import ToolCall
+            >>> env = HotPotQAEnv(
+            ...     question_id=UUID("00000000-5a8d-7341-5542-99441c6b9fe5"),
+            ...     question=(
+            ...         "Musician and satirist Allie Goertz wrote a song about the"
+            ...         ' "The Simpsons" character Milhouse, who Matt Groening named after who?'
+            ...     ),
+            ...     correct_answer="President Richard Nixon",
+            ... )
             >>> action = ToolRequestMessage(
-            ...     tool_calls=[
-            ...         ToolCall(
-            ...             function=ToolCallFunction(name="Search"),
-            ...             arguments={"entity": "Python"},
-            ...         )
-            ...     ]
+            ...     tool_calls=[ToolCall.from_name("search", entity="Python")]
             ... )
             >>> obs, done = await env.step(action)
             >>> print(obs, done)
@@ -309,23 +327,20 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
 
         # We accumulate reward across all tool calls in a step
         self.state.reward = 0.0
-        valid_action, invalid_action = self.filter_invalid_tool_calls(action)
-        # NOTE: valid_action or invalid_action may have an empty list of tool calls
         response_messages = cast(
-            list[Message],
-            # Ordered since things like search -> lookup need to be run in order.
+            "Messages",
+            # Non-concurrent since things like search -> lookup need to be run in order.
             # NOTE: Handling tool exceptions here keeps the trajectory going, but I don't
             # think the returned message is useful to the agent/learning. Disabling for now.
-            await self.exec_tool_calls(
-                valid_action, ordered=True, handle_tool_exc=False
-            )
-            + [
-                ToolResponseMessage.from_call(tool_call, content="Invalid tool call.")
-                for tool_call in invalid_action.tool_calls
-            ],
+            await self.exec_tool_calls(action, concurrency=False),
         )
         return response_messages, self.state.reward, self.state.done, False
 
+    async def get_id(self) -> str:
+        if self.question_id is None:
+            raise ValueError("No question ID was configured.")
+        return str(self.question_id)
+
     def export_frame(self) -> Frame:
         """Export the current state of the environment as a Frame object.
 
@@ -345,13 +360,22 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
                 - "answer" (Optional[str]): The answer provided by the agent, if any.
 
         Example:
-            >>> env = HotPotQAEnv()
+            >>> env = HotPotQAEnv(
+            ...     question_id=UUID("00000000-5a8d-7341-5542-99441c6b9fe5"),
+            ...     question=(
+            ...         "Musician and satirist Allie Goertz wrote a song about the"
+            ...         ' "The Simpsons" character Milhouse, who Matt Groening named after who?'
+            ...     ),
+            ...     correct_answer="President Richard Nixon",
+            ... )
             >>> frame = env.export_frame()
             >>> print(frame.info)
             {'steps': 0, 'done': False, 'reward': 0.0, 'answer': None}
         """
         return Frame(
             info={
+                "question_id": self.question_id,
+                "question": self.question,
                 "steps": self.state.steps,
                 "done": self.state.done,
                 "reward": self.state.reward,
@@ -359,7 +383,7 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
             }
         )
 
-    async def finish(self, answer: str) -> str:
+    async def submit_answer(self, answer: str) -> str:
         """Finish the task by submitting an answer to the question.
 
         Args:
@@ -367,7 +391,7 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
         """
         self.state.done = True
         if not answer:
-            return "Finish failed. No answer provided."
+            return "submit_answer failed since it was provided an empty answer."
 
         self.state.answer = answer
         self.state.reward += await self.calculate_answer_reward(answer)
@@ -401,7 +425,7 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
             response_text = self.wiki_cache[search_entity]
         except KeyError:
             # follow_redirects=True because wikipedia frequently redirects to the correct page
-            async with httpx.AsyncClient(
+            async with httpx_aiohttp.HttpxAiohttpClient(
                 follow_redirects=True, proxy=self.proxy
             ) as client:
                 response = await fetch_with_retry(
@@ -417,7 +441,10 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
         if result_divs:  # mismatch
             result_titles = [clean_str(div.get_text().strip()) for div in result_divs]
             self.state.page = None
-            return f"Could not find {entity}. Similar: {result_titles[:5]}."
+            return (
+                f"Search for {entity!r} returned no results. Similar:"
+                f" {result_titles[:5]}."
+            )
 
         page = [p.get_text().strip() for p in soup.find_all("p") + soup.find_all("ul")]
         if any("may refer to:" in p for p in page):  # Recurse
@@ -440,7 +467,7 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
         self.state.last_action_is_lookup = False
         return " ".join(obs_list[:5])
 
-    def construct_lookup_list(self, keyword: str) -> str:
+    def lookup(self, keyword: str) -> str:
         """Construct a list of sentences from the given page content that contain the specified keyword.
 
         Args:
@@ -458,7 +485,7 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
             Kearney Zzyzwicz. Milhouse has a crush on Bart's sister, Lisa, a common plot element.")
 
             keyword = "Milhouse"
-            result = construct_lookup_list_from_page(keyword, page_content)
+            result = lookup(keyword, page_content)
             # Output: ['Milhouse Mussolini Van Houten is a recurring character in the Fox animated television series The
             Simpsons.',
             'Milhouse is Bart Simpson's best friend in Mrs. Krabappel's fourth grade class at Springfield Elementary
@@ -466,7 +493,6 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
             'Milhouse is a regular target for school bully Nelson Muntz and his friends Jimbo Jones, Dolph Starbeam and
             Kearney Zzyzwicz.',
             'Milhouse has a crush on Bart's sister, Lisa, a common plot element.']
-
         """
         if not keyword:
             self.state.done = True
@@ -510,6 +536,7 @@ class HotPotQAEnvConfig(BaseModel):
     """
 
     model_config = ConfigDict(extra="forbid")
+    PROXY_ENVIRON_VAR_NAME: ClassVar[str] = "AVIARY_ENV_HOTPOTQA_PROXY"
 
     shuffle_data: bool = Field(
         default=False, description="Set True to shuffle the dataset after loading."
@@ -535,23 +562,39 @@ class HotPotQAEnvConfig(BaseModel):
             "Note that not all splits have a 'level' field."
         ),
     )
-    proxy: str | None = None
+    proxy: str | None = Field(
+        default=None,
+        description=(
+            "Optional proxy URL to use, default is unused. If unspecified, the"
+            f" environment variable {PROXY_ENVIRON_VAR_NAME} will be checked."
+        ),
+    )
+
+    @field_validator("proxy")
+    @classmethod
+    def check_proxy_environ_var(cls, value: str | None) -> str | None:
+        if not value:
+            value = os.environ.get(cls.PROXY_ENVIRON_VAR_NAME, "") or None
+        return value
+
+    evaluation_mode: EvalAnswerMode = EvalAnswerMode.CONTAINS
 
 
 class HotPotQADataset(TaskDataset[HotPotQAEnv]):
     # SEE: https://huggingface.co/datasets/hotpotqa/hotpot_qa
     HOTPOTQA_HUGGING_FACE_DATASET = "hotpotqa/hotpot_qa"
 
-    def get_data_from_hugging_face(
-        self, split: str, hf_dataset: str = HOTPOTQA_HUGGING_FACE_DATASET
-    ) -> list[tuple[str, str]]:
-        """Convert a local file and split to a list of (question, answer) tuples."""
+    @staticmethod
+    def load_raw(
+        split: str, hf_dataset: str = HOTPOTQA_HUGGING_FACE_DATASET
+    ) -> "Dataset":
+        """Load the HotPotQA dataset split from Hugging Face."""
         if split in {"dev", "eval", "val"}:  # Map common aliases
             split = "validation"
         all_datasets = load_dataset(hf_dataset, name="fullwiki", trust_remote_code=True)
         try:
-            data = all_datasets[split].select_columns(
-                column_names=["question", "answer", "level"]
+            return all_datasets[split].select_columns(
+                column_names=["id", "question", "answer", "level"]
             )
         except KeyError as exc:
             raise ValueError(
@@ -559,15 +602,33 @@ class HotPotQADataset(TaskDataset[HotPotQAEnv]):
                 f" please specify a split from {set(all_datasets.keys())}."
             ) from exc
 
+    def get_data_from_hugging_face(
+        self, split: str, hf_dataset: str = HOTPOTQA_HUGGING_FACE_DATASET
+    ) -> list[tuple[UUID, str, str]]:
+        """Get a list of (id, question, answer) tuples for the Hugging Face dataset."""
+        data = self.load_raw(split, hf_dataset)
+
         if not all(  # Making up for datasets not being typed: https://github.com/huggingface/datasets/issues/3841
-            isinstance(d["question"], str) and isinstance(d["answer"], str | float)
+            isinstance(d["id"], str)
+            and isinstance(d["question"], str)
+            and isinstance(d["answer"], str | float)
             for d in data
         ):
             raise ValueError(
                 f"Dataset {hf_dataset!r} and split {split!r} contains invalid"
-                " question(s) or answer(s)."
+                " ID(s), question(s), or answer(s)."
             )
-        return [(d["question"], d["answer"]) for d in data if self._filter_task(d)]
+        return [
+            (
+                # Sadly, using UUID v4 doesn't work with left padding (has collisions)
+                # or right padding (is a lossy conversion), so leave version unspecified
+                UUID(bytes=b"\x00" * 4 + bytes.fromhex(d["id"])),  # Left zero-pad
+                d["question"],
+                d["answer"],
+            )
+            for d in data
+            if self._filter_task(d)
+        ]
 
     def __init__(
         self, split: str, config: HotPotQAEnvConfig | dict | None = None, **kwargs
@@ -604,6 +665,7 @@ class HotPotQADataset(TaskDataset[HotPotQAEnv]):
             correct_reward=self.config.correct_reward,
             incorrect_reward=self.config.incorrect_reward,
             tool_failure_reward=self.config.tool_failure_reward,
+            evaluation_mode=self.config.evaluation_mode,
             proxy=self.config.proxy,
         )
 
