@@ -5,11 +5,12 @@ import importlib
 import inspect
 import json
 import logging
+import os
 import random
 from abc import ABC, abstractmethod
-from collections.abc import Iterator
+from collections.abc import Awaitable, Iterator
 from copy import deepcopy
-from typing import Annotated, Generic, Self, TypeAlias, TypeVar, cast
+from typing import Annotated, ClassVar, Generic, Self, TypeAlias, TypeVar, cast
 
 from pydantic import (
     BaseModel,
@@ -22,8 +23,14 @@ from pydantic import (
 )
 
 from aviary.message import Message
-from aviary.tools import Tool, ToolCall, ToolRequestMessage, ToolResponseMessage
-from aviary.utils import is_coroutine_callable
+from aviary.tools import (
+    Messages,
+    Tool,
+    ToolCall,
+    ToolRequestMessage,
+    ToolResponseMessage,
+)
+from aviary.utils import format_exc, is_coroutine_callable
 
 logger = logging.getLogger(__name__)
 
@@ -33,6 +40,16 @@ logger = logging.getLogger(__name__)
 Serializable: TypeAlias = dict | list | int | float | str | bool | BaseModel
 
 
+_T = TypeVar("_T")
+
+
+async def maybe_wait_for(future: Awaitable[_T], timeout: float | None) -> _T:
+    """Apply a timeout to an awaitable if one is provided, otherwise await indefinitely."""
+    if timeout is None:
+        return await future
+    return await asyncio.wait_for(future, timeout)
+
+
 class Frame(BaseModel):
     """A frame is a snapshot at a given timestep. The name comes from video frame."""
 
@@ -45,7 +62,7 @@ class Frame(BaseModel):
     )
 
     @staticmethod
-    def _custom_serializer(value: Serializable, handler, info):  # noqa: ARG004
+    def _custom_serializer(value: Serializable | None, handler, info):  # noqa: ARG004
         if isinstance(value, BaseModel):
             return value.model_dump()
         return handler(value)
@@ -63,7 +80,9 @@ class Frame(BaseModel):
 
     @field_validator("state", "info")
     @classmethod
-    def make_deepcopy(cls, v: Serializable, info: ValidationInfo) -> Serializable:
+    def make_deepcopy(
+        cls, v: Serializable | None, info: ValidationInfo
+    ) -> Serializable | None:
         if info.data["deepcopy"]:
             return deepcopy(v)
         return v
@@ -88,7 +107,7 @@ class Environment(ABC, Generic[TEnvState]):
     @abstractmethod
     async def step(
         self, action: ToolRequestMessage
-    ) -> tuple[list[Message], float, bool, bool]:
+    ) -> tuple[Messages, float, bool, bool]:
         """Take a step in the environment.
 
         Args:
@@ -101,7 +120,7 @@ class Environment(ABC, Generic[TEnvState]):
         """
 
     @abstractmethod
-    async def reset(self) -> tuple[list[Message], list[Tool]]:
+    async def reset(self) -> tuple[Messages, list[Tool]]:
         """
         Reset the environment and collect initial observation(s).
 
@@ -112,6 +131,23 @@ class Environment(ABC, Generic[TEnvState]):
             Two-tuple of initial observations and tools.
         """
 
+    async def get_id(self) -> str:
+        """
+        Get an identifier for this environment.
+
+        The main use case is something like the ID of the task from a dataset.
+        Since datasets may not enforce uniqueness in their IDs, we cannot ensure the IDs
+        returned from this method are unique either.
+
+        The return should not be affected by state, in other words across reset/step,
+        the return value should not change.
+
+        This method is asynchronous to allow for DB transactions.
+        """
+        raise NotImplementedError(
+            f"Getting ID is not yet implemented for environment {type(self).__name__}."
+        )
+
     def filter_invalid_tool_calls(
         self, message: ToolRequestMessage
     ) -> tuple[ToolRequestMessage, ToolRequestMessage]:
@@ -134,7 +170,7 @@ class Environment(ABC, Generic[TEnvState]):
             else:
                 invalid.append(tool_call)
         return cast(
-            tuple[ToolRequestMessage, ToolRequestMessage],
+            "tuple[ToolRequestMessage, ToolRequestMessage]",
             tuple(
                 ToolRequestMessage(
                     role=message.role,
@@ -146,11 +182,17 @@ class Environment(ABC, Generic[TEnvState]):
             ),
         )
 
+    TOOL_CALL_EXC_LOG_TB_NONDEBUG: ClassVar[bool] = os.environ.get(
+        "AVIARY_TOOL_CALL_EXC_LOG_TB_NONDEBUG", "false"
+    ).lower() in {"1", "true", "yes", "on"}
+
     async def exec_tool_calls(
         self,
         message: ToolRequestMessage,
-        ordered: bool = False,
+        concurrency: bool = True,
         handle_tool_exc: bool = False,
+        handle_invalid_tool_calls: bool = True,
+        exec_timeout: float | None = None,
         **function_kwargs,
     ) -> list[ToolResponseMessage]:
         """
@@ -158,10 +200,14 @@ class Environment(ABC, Generic[TEnvState]):
 
         Args:
             message: ToolRequestMessage containing the tool calls.
-            ordered: Opt-in flag for forcing sequential execution (according to order
-                in the above message), otherwise tool calls are made concurrently.
+            concurrency: Flag to set True (default) to concurrently execute tool calls,
+                otherwise set False to execute tool calls in the provided order.
             handle_tool_exc: Opt-in flag to suppress Exceptions and return them as a
                 ToolResponseMessage.
+            handle_invalid_tool_calls: Flag to handle invalid tool calls by returning
+                a ToolResponseMessage with a note that the tool requested doesn't exist
+            exec_timeout: Timeout for each tool call in seconds. If None, no timeout.
+                Note that handle_tool_exec can be used to catch TimeoutErrors.
             **function_kwargs: Keyword arguments to pass to all tool functions.
 
         Returns:
@@ -194,32 +240,43 @@ class Environment(ABC, Generic[TEnvState]):
             tool_exc: Exception | None = None
             try:
                 if is_coroutine_callable(tool._tool_fn):
-                    content = await tool._tool_fn(
-                        **tool_call.function.arguments, **filtered_kwargs
+                    content = await maybe_wait_for(
+                        tool._tool_fn(
+                            **tool_call.function.arguments, **filtered_kwargs
+                        ),
+                        exec_timeout,
                     )
                 else:
                     # If the function is synchronous, run on a thread
-                    content = await asyncio.to_thread(
-                        tool._tool_fn,
-                        **tool_call.function.arguments,
-                        **filtered_kwargs,
+                    content = await maybe_wait_for(
+                        asyncio.to_thread(
+                            tool._tool_fn,
+                            **tool_call.function.arguments,
+                            **filtered_kwargs,
+                        ),
+                        exec_timeout,
                     )
             except Exception as exc:
                 if not handle_tool_exc:
                     raise
                 logger_msg = (
-                    f"Encountered exception during tool call for tool {tool.info.name}:"
-                    f" {exc!r}"
+                    f"Encountered exception during tool call"
+                    f" for tool {tool.info.name}: {format_exc(exc)}"
                 )
-                # logger.exception is just too verbose and clogs up console logging. This is a
-                # more human-friendly version: log a readable error message and emit the exception
-                # at DEBUG level.
-                logger.error(logger_msg)  # noqa: TRY400
-                logger.debug(str(exc), exc_info=True)
+                if self.TOOL_CALL_EXC_LOG_TB_NONDEBUG:
+                    logger.exception(logger_msg)
+                else:
+                    # logger.exception is just too verbose and clogs up console logging.
+                    # This is a more human-friendly version:
+                    # log a readable error message and emit the exception at DEBUG level.
+                    logger.error(logger_msg)  # noqa: TRY400
+                    logger.debug(str(exc), exc_info=True)
                 tool_exc = exc
             if tool_exc:
                 # No need to mention tool.info.name here, since it'll get wrapped in a ToolResponseMessage
-                s_content = f"Encountered exception during tool call: {tool_exc}"
+                s_content = (
+                    f"Encountered exception during tool call: {format_exc(tool_exc)}"
+                )
             elif isinstance(content, str):
                 s_content = content
             elif isinstance(content, BaseModel):
@@ -228,11 +285,30 @@ class Environment(ABC, Generic[TEnvState]):
                 s_content = json.dumps(content)
             return ToolResponseMessage.from_call(tool_call, content=s_content)
 
-        if not ordered:
-            return await asyncio.gather(
-                *(_exec_tool_call(tool_call) for tool_call in message.tool_calls)
+        invalid_responses = []
+        valid_action = message
+        call_ordering = [t.id for t in message.tool_calls]
+        if handle_invalid_tool_calls:
+            valid_action, invalid_action = self.filter_invalid_tool_calls(message)
+            invalid_responses = [
+                ToolResponseMessage.from_call(
+                    tool_call, content=f"Invalid tool call: {tool_call.function.name}"
+                )
+                for tool_call in invalid_action.tool_calls
+            ]
+
+        if concurrency:
+            valid_responses = await asyncio.gather(
+                *(_exec_tool_call(tc) for tc in valid_action.tool_calls)
             )
-        return [await _exec_tool_call(tool_call) for tool_call in message.tool_calls]
+        else:
+            valid_responses = [
+                await _exec_tool_call(tc) for tc in valid_action.tool_calls
+            ]
+        return sorted(
+            invalid_responses + valid_responses,
+            key=lambda x: call_ordering.index(x.tool_call_id),
+        )
 
     def export_frame(self) -> Frame:
         """
@@ -393,7 +469,7 @@ class TaskConfig(BaseModel):
 
 
 class DummyEnvState(BaseModel):
-    messages: list[Message]
+    messages: Messages
     reward: float = 0
     done: bool = False
 
@@ -403,9 +479,20 @@ class DummyEnv(Environment[DummyEnvState]):
 
     State = DummyEnvState
 
-    def __init__(self, task: str | None = None, end_immediately: bool = True):
+    def __init__(
+        self,
+        task: str | None = None,
+        end_immediately: bool = True,
+        concurrent_tool_calls: bool = True,
+    ):
         self.end_immediately = end_immediately
         self.task = task
+        self.concurrent_tool_calls = concurrent_tool_calls
+
+    async def get_id(self) -> str:
+        if self.task is None:
+            raise ValueError("No task (to reuse as an ID) was configured.")
+        return self.task
 
     @classmethod
     def from_task(cls, task: str) -> DummyEnv:
@@ -413,12 +500,14 @@ class DummyEnv(Environment[DummyEnvState]):
 
     async def step(
         self, action: ToolRequestMessage
-    ) -> tuple[list[Message], float, bool, bool]:
-        msgs: list[Message] = await self.exec_tool_calls(action, state=self.state)
+    ) -> tuple[Messages, float, bool, bool]:
+        msgs: Messages = await self.exec_tool_calls(  # type: ignore[assignment]
+            action, state=self.state, concurrency=self.concurrent_tool_calls
+        ) or [Message(content=f"No tool calls input in tool request {action}.")]  # type: ignore[list-item]
         self.state.messages.extend(msgs)
         return msgs, self.state.reward, self.state.done, False
 
-    async def reset(self) -> tuple[list[Message], list[Tool]]:
+    async def reset(self) -> tuple[Messages, list[Tool]]:
         def print_story(story: str, state: DummyEnvState) -> None:  # noqa: ARG001
             """Print a story.
 
@@ -426,7 +515,7 @@ class DummyEnv(Environment[DummyEnvState]):
                 story: Story to print.
                 state: Environment state.
             """
-            state.reward = 1
+            state.reward = 1.0
             state.done = self.end_immediately
 
         def cast_float(x: str) -> float:
@@ -437,10 +526,15 @@ class DummyEnv(Environment[DummyEnvState]):
             """Cast the input argument x to an integer."""
             return int(x)
 
+        def get_random_int() -> int:
+            """Get a random integer in 1 to 10."""
+            return random.randint(1, 10)
+
         self.tools = [
             Tool.from_function(print_story),
             Tool.from_function(cast_float, allow_empty_param_descriptions=True),
             Tool.from_function(cast_int, allow_empty_param_descriptions=True),
+            Tool.from_function(get_random_int, allow_empty_param_descriptions=True),
         ]
         self.state = type(self).State(
             messages=[
