@@ -10,6 +10,7 @@ import pytest
 from fastapi.testclient import TestClient
 from pydantic import BaseModel, Field
 from pytest_subtests import SubTests
+from tenacity import retry, retry_if_exception_type, stop_after_attempt
 from typeguard import suppress_type_checks
 
 from aviary.core import (
@@ -58,6 +59,7 @@ def many_edge_cases(
     enum: StubEnum,
     defaulted_str: str = "default",
     defaulted_float: float = 1.0,
+    structured_arg: str = "structured",
 ) -> None:
     """
     Check using docstrings as partial f-string templates like so: {summary_format}.
@@ -73,6 +75,22 @@ def many_edge_cases(
         enum: I am an enum.
         defaulted_str: I have a string default value.
         defaulted_float: I have a float default value.
+        structured_arg: I am structured. There are lots of examples
+            included which cross several lines.
+            Query Syntax:
+                Basic Search:
+                    Search with two words and a space:
+                    >>> "word1 word2"  # doctest: +SKIP
+
+                Modified Search:
+                    Use operators to modify search behavior:
+                    >>> 'EXPANSION[None]NAME"word phrase"'  # doctest: +SKIP
+                    >>> 'EXPANSION[Concept]"words"'  # doctest: +SKIP
+
+            Operators:
+                EXPANSION[type]: Terms
+                    - Term1: a description
+                    - Term2: another description
     """
 
 
@@ -108,6 +126,33 @@ def example_fxn(x: int, y: str, z: float) -> None:
     assert isinstance(z, float)
 
 
+def state_out_of_order(state: dict, x: int = 0) -> None:
+    """A test function.
+
+    There may be non-summary content.
+
+    Args:
+        state: this should not be present
+        x: x
+    """
+    assert isinstance(state, dict)
+    assert isinstance(x, int)
+
+
+@retry(retry=retry_if_exception_type(ValueError), stop=stop_after_attempt(3))
+def add_with_tenacity_retries(a: int, b: int | None = 0) -> int:
+    """Add two numbers.
+
+    Args:
+        a: first number
+        b: second number
+    """
+    if b is None:
+        # Intentionally don't use ValueError since we don't want to trigger retries
+        raise TypeError("Please pass b as a number.")
+    return a + b
+
+
 class TestTool:
     @pytest.mark.parametrize(
         ("fn", "kwargs", "expected"),
@@ -169,7 +214,7 @@ class TestTool:
                                 },
                                 "y": {
                                     "description": (
-                                        "I am null. And despite that there is a"
+                                        "I am null.\nAnd despite that there is a"
                                         " multiline argument description."
                                     ),
                                     "title": "Y",
@@ -226,6 +271,29 @@ class TestTool:
                                     "title": "Defaulted Float",
                                     "type": "number",
                                 },
+                                "structured_arg": {
+                                    "default": "structured",
+                                    "description": (
+                                        "I am structured. There are lots of examples\n"
+                                        "included which cross several lines.\n"
+                                        "Query Syntax:\n"
+                                        "    Basic Search:\n"
+                                        "        Search with two words and a space:\n"
+                                        '        >>> "word1 word2"  # doctest: +SKIP\n'
+                                        "\n"
+                                        "    Modified Search:\n"
+                                        "        Use operators to modify search behavior:\n"
+                                        "        >>> 'EXPANSION[None]NAME\"word phrase\"'  # doctest: +SKIP\n"
+                                        "        >>> 'EXPANSION[Concept]\"words\"'  # doctest: +SKIP\n"
+                                        "\n"
+                                        "Operators:\n"
+                                        "    EXPANSION[type]: Terms\n"
+                                        "        - Term1: a description\n"
+                                        "        - Term2: another description"
+                                    ),
+                                    "title": "Structured Arg",
+                                    "type": "string",
+                                },
                             },
                             "required": [
                                 "x",
@@ -296,6 +364,32 @@ class TestTool:
                 },
                 id="with-args-and-returns",
             ),
+            pytest.param(
+                state_out_of_order,
+                {},
+                {
+                    "type": "function",
+                    "info": {
+                        "name": "state_out_of_order",
+                        "description": (
+                            "A test function.\n\nThere may be non-summary content."
+                        ),
+                        "parameters": {
+                            "properties": {
+                                "x": {
+                                    "default": 0,
+                                    "description": "x",
+                                    "title": "X",
+                                    "type": "integer",
+                                },
+                            },
+                            "required": [],
+                            "type": "object",
+                        },
+                    },
+                },
+                id="with-state-out-of-order",
+            ),
             pytest.param(
                 example_fxn,
                 {},
@@ -331,6 +425,37 @@ class TestTool:
                 },
                 id="with-linefeed",
             ),
+            pytest.param(
+                # NOTE: tenacity uses `functools.wraps`: https://github.com/jd/tenacity/blob/9.1.2/tenacity/__init__.py#L330-L333
+                # Also, we commonly use `tenacity.retry`, so it's useful to check this too
+                add_with_tenacity_retries,
+                {},
+                {
+                    "type": "function",
+                    "info": {
+                        "name": "add_with_tenacity_retries",
+                        "description": "Add two numbers.",
+                        "parameters": {
+                            "properties": {
+                                "a": {
+                                    "description": "first number",
+                                    "title": "A",
+                                    "type": "integer",
+                                },
+                                "b": {
+                                    "description": "second number",
+                                    "title": "B",
+                                    "default": 0,
+                                    "anyOf": [{"type": "integer"}, {"type": "null"}],
+                                },
+                            },
+                            "required": ["a"],
+                            "type": "object",
+                        },
+                    },
+                },
+                id="functools-wraps-decoration",
+            ),
         ],
     )
     def test_from_function(
@@ -385,7 +510,7 @@ PARAMETERS:
 
     def test_describe(self, subtests: SubTests) -> None:
         """Test that describe_xyz functions for FunctionInfo are reasonable."""
-        tool = Tool.from_function(example_fxn)
+        tool = Tool.from_function(many_edge_cases)
 
         with subtests.test("Test describe_xml is callable"):
             assert tool.info.describe_xml()
@@ -393,6 +518,9 @@ PARAMETERS:
         with subtests.test("Test describe_json is callable"):
             assert tool.info.describe_json()
 
+        with subtests.test("Test describe_str is callable"):
+            assert tool.info.describe_str()
+
     def test_serialization_manual(self) -> None:
         # make one manually
         tool = Tool(
@@ -460,6 +588,7 @@ PARAMETERS:
     async def test_arg_types(self) -> None:
         tool = Tool.from_function(example_fxn)
 
+        assert tool.info.parameters is not None
         assert tool.info.parameters.properties["x"]["type"] == "integer"
         assert tool.info.parameters.properties["y"]["type"] == "string"
         assert tool.info.parameters.properties["z"]["type"] == "number"
@@ -561,8 +690,7 @@ PARAMETERS:
             assert new_messages[0].content == "Go for a walk"
 
 
-@pytest.mark.asyncio
-async def test_argref_by_name_basic_usage() -> None:
+def test_argref_by_name_basic_usage() -> None:
     class MyState:
         def __init__(self):
             self.refs = {"foo": 1}
@@ -582,8 +710,7 @@ async def test_argref_by_name_basic_usage() -> None:
     assert s.refs[name] == 1 + 2
 
 
-@pytest.mark.asyncio
-async def test_argref_by_name_error_handling() -> None:
+def test_argref_by_name_error_handling() -> None:
     class MyState:
         def __init__(self):
             self.refs = {"foo": 1}
@@ -612,7 +739,7 @@ async def test_argref_by_name_async_functions() -> None:
 
     # Define the async_add function with the decorator
     @argref_by_name()
-    async def async_add(a: int, b: int) -> int:
+    async def async_add(a: int, b: int) -> int:  # noqa: RUF029
         """Some docstring."""
         return a + b
 
@@ -633,7 +760,7 @@ async def test_argref_by_name_async_functions() -> None:
 
     # Define and test async_list
     @argref_by_name()
-    async def async_list(a: int, b: int) -> list[int]:
+    async def async_list(a: int, b: int) -> list[int]:  # noqa: RUF029
         """Some docstring."""
         return [a, b]
 
@@ -644,7 +771,7 @@ async def test_argref_by_name_async_functions() -> None:
 
     # Define and test async_list_direct
     @argref_by_name(return_direct=True)
-    async def async_list_direct(a: int, b: int) -> list[int]:
+    async def async_list_direct(a: int, b: int) -> list[int]:  # noqa: RUF029
         """Some docstring."""
         return [a, b]
 
@@ -685,7 +812,7 @@ async def test_argref_by_name_advanced_features() -> None:
 
     # Test state passing with fxn_requires_state
     @argref_by_name(fxn_requires_state=True)
-    async def want_state(a: int, state: MyState) -> int:  # noqa: ARG001
+    async def want_state(a: int, state: MyState) -> int:  # noqa: ARG001, RUF029
         """Some docstring.
 
         Args:
@@ -708,8 +835,7 @@ async def test_argref_by_name_advanced_features() -> None:
     assert kwarg_list_test(a="foo,foo", state=s) == 2
 
 
-@pytest.mark.asyncio
-async def test_argref_by_name_type_checking() -> None:
+def test_argref_by_name_type_checking() -> None:
     class MyInt(int):
         pass
 
